#include <stdio.h>

/*
//---------------------------------------------------------------------------
// defines
//---------------------------------------------------------------------------
*/

#define TUNNEL_LENGTH 10

/*
//---------------------------------------------------------------------------
   types
//---------------------------------------------------------------------------
*/

typedef signed short  INT16;  /* signed 16 bit integer              */
typedef unsigned int  UINTW;  /* fastest integer (16 bits or above) */
typedef signed short  UINT16; /* unsigned 16 bit integer            */
typedef unsigned long UINT32; /* unsigned 32 bit integer            */

/*
//---------------------------------------------------------------------------
// structs
//---------------------------------------------------------------------------
*/

typedef struct {
  double x,y,z;
} VERT;

typedef struct {
  INT16  u,v,l;
} UVL;

typedef struct {
  UINT16 texture1;
  UINT16 texture2;
  UVL    uvl[4];
} SIDE;

typedef struct {
  SIDE   side[6];
  UINT16 child[6];
  VERT   vert[8];
  INT16  light;
} SEGMENT;

/*
//---------------------------------------------------------------------------
// prototypes
//---------------------------------------------------------------------------
*/

void dump_seg(FILE *file,UINTW segnum,SEGMENT *seg);


/*
//---------------------------------------------------------------------------
// constants
//---------------------------------------------------------------------------
*/


const UVL default_uvl[4] = {
  { 0x0000, 0x0000,0x8000},
  { 0x0000, 0x0800,0x8000},
  {-0x0800, 0x0800,0x8000},
  {-0x0800, 0x0000,0x8000}
};

/*
// 7o-----o4       Y
//  |\    |\       ^
//  | \   | \      |
//  | 3o-----o0    |
//  |  |  |  |     |
// 6o--|--o5 |     0---->X
//   \ |   \ |      \
//    \|    \|       \
//    2o-----o1       \Z
//
*/


const VERT standard_vert[8] = {
  {  0,  0,  0},
  {  0,-20,  0},
  {-20,-20,  0},
  {-20,  0,  0},
  {  0,  0,-20},
  {  0,-20,-20},
  {-20,-20,-20},
  {-20,  0,-20}
};

const char title[] = "BLKGEN v1.0 - Block Generator Copyright (C) 1997"
                     " Bryan Aamot, Brainware\n";
const char usage[] = "Generates a BLK file for the Descent Mission Builder\n"
                     "Usage: BLKGEN <filename>\n"
                     "  will generate a sample block file\n"
                     "Example: BLKGEN test.blk\n";

/*
//---------------------------------------------------------------------------
// globals
//---------------------------------------------------------------------------
*/


/*
//---------------------------------------------------------------------------
// main
//---------------------------------------------------------------------------
*/

int main (int ac, char *av[]) {
  SEGMENT segment;
  SEGMENT *seg = &segment;
  UINTW i,j,cube;
  FILE *file = NULL;
  char *filename = av[1];

  printf(title);

  if (ac != 2) {
    printf(usage);
    goto abort;
  }

  file = fopen(filename,"wt");
  if (!file) {
    printf("Could not open file '%s'\n",filename);
    goto abort;
  } else {
    fprintf(file,"DMB_BLOCK_FILE\n");

    /*
    // For this example, we will use all default textures
    //
    // Note1: you must add 0x8000 to texture1 if you plan on using
    // texture 2
    //
    // Note2: sides which have children are allowed to have
    // UVL values of all 0, but it doesn't hurt to just set them
    // as the default values.
    //
    // Note3: the default UV values will "stretch to fit" the texture
    // on the cube's side.
    */

    for (i=0;i<6;i++) {
      seg->side[i].texture1 = 0;
      seg->side[i].texture2 = 0;
      for (j=0;j<4;j++) {
        seg->side[i].uvl[j].u = default_uvl[j].u;
        seg->side[i].uvl[j].v = default_uvl[j].v;
        seg->side[i].uvl[j].l = default_uvl[j].l;
      }
    }

    /*
    // This sets up which cubes connect to this cube
    // -1 = no cube, otherwise use the cube's number
    // For now, set them all to have no children.
    */

    for (i=0;i<6;i++) {
      seg->child[i] = -1;
    }

    /*
    // The points (or vertices) of the cube are based
    // on a standard cube.
    */

    for (i=0;i<8;i++) {
      seg->vert[i].x = standard_vert[i].x;
      seg->vert[i].y = standard_vert[i].y;
      seg->vert[i].z = standard_vert[i].z;
    }


    /*
    // Now that we have set up our standard cube,
    // we will modify it for our example.  In this
    // example, we will create a series of cubes
    // connected together to form a tunnel.
    // Side 4 (the front of the cube) will be
    // connected to side 5 (the back of the cube).
    // Therefore, we need to set the children
    // of the cubes to connect them.  Cube 0
    // only has one child (cube 1).  The last cube
    // (call it cube N) also only has one child
    // (cube N-1).  The rest of the cubes have
    // two children (the one in front, and the one
    // behind it).
    //
    // Each cube will be 20 units less than the
    // last cube on the Z axis.
    */

    for (cube=0;cube<TUNNEL_LENGTH;cube++) {

      /*
      // Set up child relationships
      */

      seg->child[4] = cube+1;
      seg->child[5] = cube-1;

      if (cube == 0) {
        seg->child[5] = -1;
        /*
        // Note: just so happens to be -1 already
        // since cube-1 = -1 for cube 0, but
        // leaving this step out would be kind
        // of tricky :)
        */
      }
      if (cube == TUNNEL_LENGTH - 1) {
        seg->child[4] = -1;
      }

      /*
      // Write this cube to the block file
      */

      dump_seg(file,cube,seg);

      /*
      // Move the next cube -20 on the z axis
      */

      for (i=0;i<8;i++) {
        seg->vert[i].z -= 20;
      }

    }
    fclose(file);
  }

  printf("%s created\n",filename);
  return 0;

abort:
  return 1;

}



/*
//---------------------------------------------------------------------------
// dump_seg()
//
// ACTION - Writes a segment's information in text form to a file.  Uses
//          actual coordinate information instead of vertix number.  Only
//          saves segment information (no walls, objects, or triggers).
//
//---------------------------------------------------------------------------
*/
void dump_seg(FILE *file,UINTW segnum, SEGMENT *seg) {
  UINTW i,j;

  fprintf(file,"segment %d\n",segnum);
  for (i=0;i<6;i++) {
    fprintf(file,"  side %d\n",i);
    fprintf(file,"    tmap_num %d\n",seg->side[i].texture1);
    fprintf(file,"    tmap_num2 %d\n",seg->side[i].texture2);
    for (j=0;j<4;j++) {
      fprintf(file,"    uvls %d %d %d\n",seg->side[i].uvl[j].u,
					 seg->side[i].uvl[j].v,
					 seg->side[i].uvl[j].l);
    }
  }
  fprintf(file,"  children %d %d %d %d %d %d\n",
		seg->child[0],seg->child[1],seg->child[2],
		seg->child[3],seg->child[4],seg->child[5]);

  for (i=0;i<8;i++) {
    UINT32 x,y,z;
    /* convert to axis system used by block format */
    x = - (UINT32)((double)0x10000L * seg->vert[i].y);
    y =   (UINT32)((double)0x10000L * seg->vert[i].z);
    z =   (UINT32)((double)0x10000L * seg->vert[i].x);
    fprintf(file,"  vms_vector %d %ld %ld %ld\n",i,x,y,z);
  }
  fprintf(file,"  static_light %ld\n",seg->light);
}


